static sreg_t physdev_op(ureg_t *args)
{
    return -EPERM; /* we don't do dom0 */
}

static sreg_t stack_switch(ureg_t *args)
{
    xen_ctxt.kernel_ss = fix_sel(args[0]);
    xen_ctxt.kernel_sp = args[1];

#if 0 == longmode
    xen_tss.ss1  = xen_ctxt.kernel_ss;
    xen_tss.esp1 = xen_ctxt.kernel_sp;
#endif
    return 0;
}

static sreg_t update_descriptor(ureg_t *args)
{
#if longmode
    uint64_t pa = args[0];
    struct descriptor_32 desc = {
	.a = args[1] & 0xffffffff,
	.b = args[1] >> 32,
    };
#else
    uint64_t pa   = args[0] | (uint64_t)args[1] << 32;
    struct descriptor_32 desc = {
	.a = args[2],
	.b = args[3],
    };
#endif
    int p, index;

    for (p = 0; p < sizeof(gdt_mfn)/sizeof(gdt_mfn[0]); p++) {
	if (addr_to_frame(pa) == gdt_mfn[p])
	    break;
    }
    if (p == sizeof(gdt_mfn)/sizeof(gdt_mfn[0]))
	return -EINVAL;
    index = addr_offset(pa) / sizeof(struct descriptor_32);
    
    fix_desc(&desc);
    xen_gdt[p * 512 + index] = desc;
    return 0;
}

static sreg_t fpu_taskswitch(ureg_t *args)
{
    if (args[0]) {
	write_cr0(X86_CR0_TS|read_cr0());
    } else {
	__asm__("clts");
    }
    return 0;
}

static sreg_t grant_table_op(ureg_t *args)
{
    struct gnttab_setup_table *st;
    struct gnttab_query_size  *qs;
    unsigned long *frames;
    int i, rc = 0;
    
    switch (args[0]) {
    case GNTTABOP_setup_table:
	st = (void*)args[1];
	printk("%s: setup_table %d\n", __FUNCTION__, st->nr_frames);
	if (st->nr_frames > GRANT_FRAMES_MAX) {
	    st->status = GNTST_general_error;
	} else {
	    grant_frames = st->nr_frames;
	    frames = st->frame_list.p;
	    for (i = 0; i < grant_frames; i++)
		frames[i] = EMU_MFN(grant_table) + i;
	    st->status = GNTST_okay;
	}
	break;
    case GNTTABOP_query_size:
	printk("%s: query_size\n", __FUNCTION__);
	qs = (void*)args[1];
	qs->nr_frames = grant_frames;
	qs->max_nr_frames = 1;
	qs->status = GNTST_okay;
	break;
    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	rc = -ENOSYS;
    }
    return rc;
}
