#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <inttypes.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <xs.h>
#include "xennerctrl.h"
#include <xen/io/xenbus.h>

#include "shared.h"
#include "list.h"
#include "daemon.h"
#include "xenbackd.h"

/* ------------------------------------------------------------- */

#define BE_ROOT "/local/domain/0/backend/netbackd"

struct netdev {
    struct xendev       xendev;  /* must be first */
};

static int termsig;

/* ------------------------------------------------------------- */

static void catchsig(int sig)
{
    termsig = sig;
}

/* ------------------------------------------------------------- */

static int net_alloc(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    return 0;
}

static int net_setup(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    return 0;
}

static int net_backend(struct xendev *xendev, char *node, char *val)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, netdev, node,
	     val ? val : "<deleted>");
    return 0;
}

static int net_frontend(struct xendev *xendev, char *node, char *val)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, netdev, node,
	     val ? val : "<deleted>");
    return 0;
}

static int net_event(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    return 0;
}

static int net_free(struct xendev *xendev)
{
    struct netdev *netdev = container_of(xendev, struct netdev, xendev);

    d1printf("%s: %p\n", __FUNCTION__, netdev);
    return 0;
}

static struct devops netdev_ops = {
    .size  = sizeof(struct netdev),
    .alloc = net_alloc,
    .setup = net_setup,
    .xs_be = net_backend,
    .xs_fe = net_frontend,
    .event = net_event,
    .free  = net_free,
};

/* ------------------------------------------------------------- */

static void usage(FILE *fp)
{
    fprintf(fp,
	    "\n"
	    "netbackd -- no usage yet, sorry\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    struct sigaction act,old;
    int maxfd, fd;
    fd_set set;
    int c;
                
    struct xs_handle *xs = NULL;

    for (;;) {
        if (-1 == (c = getopt(argc, argv, "hdp:")))
            break;
        switch (c) {
        case 'd':
	    debug++;
	    break;
	case 'p':
	    pidfile = optarg;
	    break;

        case 'h':
            usage(stdout);
            exit(0);
        default:
            usage(stderr);
            exit(1);
        }
    }

    /* setup signal handler */
    memset(&act,0,sizeof(act));
    sigemptyset(&act.sa_mask);
    act.sa_handler = catchsig;
    sigaction(SIGTERM,&act,&old);
    sigaction(SIGINT,&act,&old);

    evtchnd = xenner_evtchn_open();
    if (evtchnd < 0) {
	fprintf(stderr, "can't connect to evtchnd\n");
	exit(1);
    }

    xs = xs_daemon_open();
    if (!xs) {
	fprintf(stderr, "can't connect to xenstored\n");
	exit(1);
    }

    /* fork into background, handle pidfile */
    daemonize();

    xenstore_scan(xs, BE_ROOT, &netdev_ops);

    /* main loop */
    for (;;) {
	if (termsig)
	    break;
	
	FD_ZERO(&set);
	maxfd = 0;

	fd = xs_fileno(xs);
	FD_SET(fd, &set);
	if (maxfd < fd)
	    maxfd = fd;

	fd = xenner_evtchn_fd(evtchnd);
	FD_SET(fd, &set);
	if (maxfd < fd)
	    maxfd = fd;

	switch (select(maxfd+1, &set, NULL, NULL, NULL)) {
	case -1:
	    if (EINTR == errno)
		continue; /* termsig check */
	    perror("select");
	    break;
	case 0:
	    break;
	default:
	    break;
	}

	if (FD_ISSET(xs_fileno(xs), &set))
	    xenstore_update(xs);
	if (FD_ISSET(xenner_evtchn_fd(evtchnd), &set))
	    event_received();
    }
    return 0;
}
