#include <xen/elfnote.h>
#include <xen/elfstructs.h>

#include "list.h"

/* ------------------------------------------------------------- */

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#define container_of(ptr, type, member) ({                      \
        const typeof( ((type *)0)->member ) *__mptr = (ptr);    \
        (type *)( (char *)__mptr - offsetof(type,member) );})

/* ------------------------------------------------------------- */

#define BUFSIZE 256

struct xendev;

struct devops {
    size_t size;
    int    (*alloc)(struct xendev *xendev);
    int    (*setup)(struct xendev *xendev);
    int    (*xs_be)(struct xendev *xendev, char *node, char *val);
    int    (*xs_fe)(struct xendev *xendev, char *node, char *val);
    int    (*event)(struct xendev *xendev);
    int    (*free)(struct xendev *xendev);
};

struct xendev {
    char               *be_root;
    int                dom;
    int                dev;
    enum xenbus_state  be_state;
    enum xenbus_state  fe_state;
    char               *fe;
    char               *protocol;
    int                ring_ref;
    int                remote_port;
    int                local_port;

    void               *vmcore;
    size_t             vmcore_size;
    Elf64_Ehdr         *vm_ehdr;
    Elf64_Phdr         *vm_phdrs;
    Elf64_Shdr         *vm_shdrs;
    char               *vm_shstrtab;
    struct grant_entry *grant_table;

    struct xs_handle   *xs;
    struct devops      *ops;
    struct list_head   next;
};

/* ------------------------------------------------------------- */

extern int evtchnd;

#if 0
struct xendev *find_xendev(int dom, int dev);
struct xendev *get_xendev(int dom, int dev, struct xs_handle *xs,
			  struct devops *ops);
struct xendev *del_xendev(int dom, int dev);
#endif

int write_be_str(struct xendev *xendev, char *node, char *val);
int write_be_int(struct xendev *xendev, char *node, int ival);
char *read_be_str(struct xendev *xendev, char *node);
int read_be_int(struct xendev *xendev, char *node);
char *read_fe_str(struct xendev *xendev, char *node);
int read_fe_int(struct xendev *xendev, char *node);

int change_state_xendev(struct xendev *xendev, enum xenbus_state state);

int vmcore_map(struct xendev *xendev);
int vmcore_unmap(struct xendev *xendev);
void *vmcore_mfn2ptr(struct xendev *xendev, uint64_t mfn);
void *vmcore_ref2ptr(struct xendev *xendev, uint32_t ref);

void xenstore_scan(struct xs_handle *xs, char *devpath,
		   struct devops *ops);
void xenstore_update(struct xs_handle *xs);
void event_received(void);
